####################################################
# to find pair of C12-C13 pairs
# Usage:
#      pairList<-findPeakPairsDupl(peaklist,chargeVec,mzTolChargeVec,intRatioRangeChargeMat,intSaturated,intRatioC12_1stTag,intRC13_1stIso,intRC13_C12Vec,mzTolIso,debugfile)
#                peaklist: data frame/matrix contains scan-number[No],m/z [mz], intensity[maxo],signal_to_noise[sn],rt
#               chargeVec: mz-dist vector to defined charge-types
#                            e.g.,: c(1.00336,0.50168,0.33445,0.25084)
#                       chargeType:      1        2       3      4
#          mzTolChargeVec: mz tolerance used to define a charge-type
#                            e.g.,: c(2.5e-6,2.5e-6,2.5e-6,2.5e-6)
#  intRatioRangeChargeMat: intensity ratio range for valid 1st isotopic of C12 [could be different for different charge-types]
#                            e.g.,: matrix(c(0.15,0.75,0.15,0.75,0.15,0.75,0.15,0.75),2,4)
#                                   different cols ==> different chargeType
#            intSaturated: saturated intensity
#      intRatioC12_1stTag: the "max" intensity ratio of C12 over possible tag 1 [for test, 3 is used]
#   intRatioC13_to_1stISO: the minimum intensity ratio of valid C13 over that of the 1st isotope of C12
#                            e.g.,: c(0.5,0.5,0.5,0.5) ==> for different tag-types
#  intRatio1stIso_C13_C12: the ratio range of [intensity ratio of 1st isotope over C13 ] over [intensity ratio of 1st isotope over C12] 
#                            e.g.,: c(0.66,1.5)  ==> lev-1 peak pair
#                mzTolIso: mz tolerance used to search isotopes [more details of isotopes in findIsotope]
#                            e.g.,: same as mzTolChargeVec
#               mzTolLev3: mz tolerance used for lev3 peak fixing
#
#               debugfile: file name for code-testing, if the filename is empty [not default], debug function will be disabled
#                          default: debugfile_findpk.csv
# Update:
#       Mar 8, 2012 :  coded the function version
#                      Lev-3 peak-pairs is NOT fixed yet [need find proper charge-type and tag-type for them] 
#       Mar24, 2012 :  re-checked the code, and keep everything consistent in the functions
#       Mar31, 2012 :  use a mzTol vector for isotope searching
#                      use a new mzTol for lev3 peak fixing
#                      tag the lev0 as lev4
#              2013 : conside the saturated condition, increase the intRatioRangeChargeMat by 1.5 
####################################################
findPeakPairsDupl<-function(peaklist,chargeVec,mzTolChargeVec,intRatioRangeChargeMat,intSaturated=4.8e+6,intRatioC12_1stTag,intRatioC13_to_1stISO,intRatio1stIso_C13_C12,mzTolIso, mzTolLev3,isFT=FALSE,debugfile=NA)
{
   if (isFT==TRUE)
   {
      print("findPeakPairsDupl: FT version")
   } else {
      print("findPeakPairsDupl: TOF version")
   }
   isOnlyPartChargeTag<-TRUE; # only consider 1charge-1tag, 1chage-2tag, and 2charge-2tag

   # sort data by scan number, then mz
   peaklist.sorted <- peaklist[order(peaklist$No,peaklist$mz,decreasing=FALSE), ]; rm(peaklist)
   
   #----------------- start of setting  --------------------
   # create the charge & tag vector, different cols ==> different chargeType
   tagMat<-matrix(c(2.0067,4.0133,6.02001,8.02668,        # 1st col: charge-type 1
                    1.0034,2.0067,3.01008,4.01344,        # 2nd col: charge-type 2
                    0.6689,1.3378,2.00670,2.67560,        # 3rd col: charge-type 3
                    0.5017,1.0034,1.50504,2.00672),4,4);  # 4th col: charge-type 4
   #----------------- end of setting  --------------------
   nChargeType<-length(chargeVec);
   if (nChargeType!=ncol(tagMat))
   {
      stop("check input chargeVec: only four charge-types are supported here")
   }
   nTagType<-nrow(tagMat);
   
   # create the sample column if necessary
   if (length(which(colnames(peaklist.sorted)=="sample"))==0)
   {
     peaklist.sorted$sample <- 1;
     samples<-1;
   } else {
     peaklist.sorted$sample <- as.factor(peaklist.sorted$sample);
     samples<-levels(peaklist.sorted$sample);
   }
   
   isDebug<-FALSE;
   if (length(debugfile)!=0 && !is.na(debugfile))
   {
      isDebug<-TRUE;
   }
   pair.count<-0;
   maxLen<-floor(nrow(peaklist.sorted)/2);
   indLightVec<-vector(mode="integer",length=maxLen);
   indHeavyVec<-vector(mode="integer",length=maxLen);

   pairLevVec<-vector(mode="integer",length=maxLen);
   pairChargeVec<-vector(mode="integer",length=maxLen);
   pairTagVec<-vector(mode="integer",length=maxLen);
   pairRatioVec<-vector(mode="numeric",length=maxLen)
   intRatioRangeChargeMatST<-intRatioRangeChargeMat*0.8

   if (isDebug==TRUE)
   {
      intRC13_C12Vec<-vector(mode="numeric",length=maxLen);
      intR1stIso_C12Vec<-vector(mode="numeric",length=maxLen);
      C13TagVec<-vector(mode="integer",length=maxLen)
   }
   indOri<-c(1:nrow(peaklist.sorted));

   for (s in 1:length(samples) )
   {
       
       indexSample<-which(peaklist.sorted$sample==samples[s]);
       #currentIndexOri<-indOri[indexSample];
       mzVec<-peaklist.sorted$mz[indexSample];
       intVec<-peaklist.sorted$maxo[indexSample];
       scanNumVec<-peaklist.sorted$No[indexSample];

       # set up a vector to remember peaks already matched or discarded
       numRowC<-length(indexSample);
       matched.peaks<-vector(mode='logical', length=numRowC);

       # unique scan numbers within current sample
       scanNum<-unique(scanNumVec);
       numOfScanNum<-length(scanNum);

       # loop each scan-number to find the pairs
       for (nS in 1:numOfScanNum)
       {
           #if (isDebug==TRUE) {print(paste("scan number:",scanNum[nS]));}
           indexCurrentNo<-which(scanNumVec==scanNum[nS])
           mzVecCurrentNo<-mzVec[indexCurrentNo]
           intVecCurrentNo<-intVec[indexCurrentNo]
           nLenCurrentNo<-length(indexCurrentNo)
           m<-0
           while (m<(nLenCurrentNo-1))
           {
               m<-m+1
               if (matched.peaks[indexCurrentNo[m]]) { next; }   
               mzA<-mzVecCurrentNo[m];               # mz for C12
               intA<-intVecCurrentNo[m];             # intensity for C12
   
               # check the charge-type, need mz and int only after m
               nCurrent<-m+1;
               isUsed<-matched.peaks[indexCurrentNo[nCurrent:nLenCurrentNo]]

               if (intA>intSaturated)
               {
                  mzACTypeList<-findChargeType(mzVecCurrentNo[nCurrent:nLenCurrentNo],intVecCurrentNo[nCurrent:nLenCurrentNo], 
                                               mzA,intA,chargeVec,mzTolChargeVec, intRatioRangeChargeMatST,isUsed);
               }  else {
                  mzACTypeList<-findChargeType(mzVecCurrentNo[nCurrent:nLenCurrentNo],intVecCurrentNo[nCurrent:nLenCurrentNo], 
                                               mzA,intA,chargeVec,mzTolChargeVec, intRatioRangeChargeMat,isUsed);
               } 
               mzACType<-mzACTypeList$chargeType;
   
               if (mzACType>0)
               {
                  # ChargeType is found, then check tag-type
                  mzB <- mzACTypeList$mz;               # mz for 1st isotope
                  intB <- mzACTypeList$int;             # intensity for 1st isotope
                  indexB<-mzACTypeList$ind;             # local index of 1st isotope
                  indexBReal<-indexB+nCurrent-1;        # index in current scan number
                  nCurrent<-indexBReal+1;
                  if (nCurrent>nLenCurrentNo)
                  {
                     mzATType <- -1;
                  } else {
                     isUsed<-matched.peaks[indexCurrentNo[nCurrent:nLenCurrentNo]]
                     mzATTypeList<-findTagType(mzVecCurrentNo[nCurrent:nLenCurrentNo],intVecCurrentNo[nCurrent:nLenCurrentNo],
                                               mzA,intA,tagMat[, mzACType], mzTolChargeVec, mzB,intB,intRatioC12_1stTag,intRatioC13_to_1stISO,isUsed);
                     mzATType<-mzATTypeList$tagType;
                  }
                  
                  # 20130526
                  isTargetType<-FALSE
                  if ((mzACType==1 && mzATType==1) || (mzACType==1 && mzATType==2) || (mzACType==2 && mzATType==2) ) {isTargetType<-TRUE}
                  if (isOnlyPartChargeTag==FALSE && mzATType>0) {isTargetType<-TRUE}

                  if (isTargetType==TRUE)
                  {
                     mzC <- mzATTypeList$mz[mzATType];   # mz for C13
                     intC<-mzATTypeList$int[mzATType];   # intensity for C13
                     indexC<-mzATTypeList$ind[mzATType]; # local index of C13
                     indexCReal<-indexC+nCurrent-1;      # index in current scan number
                     nCurrent<-indexCReal+1;

                     # search 1st isotope of C13
                     if (nCurrent>nLenCurrentNo)
                     {
                       mzCCType<-0;
                     } else {
                       isUsed<-matched.peaks[indexCurrentNo[nCurrent:nLenCurrentNo]]
                       if (intC>intSaturated)
                       {
                          mzCCTypeList<-findChargeType(mzVecCurrentNo[nCurrent:nLenCurrentNo],intVecCurrentNo[nCurrent:nLenCurrentNo], 
                                                       mzC,intC,chargeVec,mzTolChargeVec, intRatioRangeChargeMatST,isUsed);
                       }  else {
                          mzCCTypeList<-findChargeType(mzVecCurrentNo[nCurrent:nLenCurrentNo],intVecCurrentNo[nCurrent:nLenCurrentNo], 
                                                       mzC,intC,chargeVec,mzTolChargeVec, intRatioRangeChargeMat,isUsed);
                       }
                       mzCCType<-mzCCTypeList$chargeType;
                     }
                     pair.count<-pair.count+1;                    
                     if (mzCCType==mzACType)      # ************ C12 & C13 must have same charge number **************
                     {
                        ratio1stIso_C12 <- intB/intA;
                        ratio1stIso_C13 <- mzCCTypeList$int/intC;
                        rr <- ratio1stIso_C12/ratio1stIso_C13;
                        if (isDebug==TRUE) { intRC13_C12Vec[pair.count] <- rr; };
                        if (rr>intRatio1stIso_C13_C12[1] && rr<intRatio1stIso_C13_C12[2])
                        {
                           pairLevVec[pair.count]<-1;
                           # check int_C13 to calculated ratio using the 1st Isotopes
                           if (intA>intSaturated)
                           {
                              if (mzA>=250 && mzA<350)
                              {
                                 magRatio<-1.1
                                 pairRatioVec[pair.count]<-intB/mzCCTypeList$int/magRatio
                              } else if (mzA>=350 && mzA<450) {
                                 magRatio<-1.05
                                 pairRatioVec[pair.count]<-intB/mzCCTypeList$int/magRatio
                              } else if (mzA>=450 && mzA<550) {
                                 magRatio<-1.05
                                 pairRatioVec[pair.count]<-intB/mzCCTypeList$int/magRatio
                              } else if (mzA>=550) {
                                 magRatio<-1.05
                                 pairRatioVec[pair.count]<-intB/mzCCTypeList$int/magRatio
                              }
                           }
                        } else {
                           # found but intensity ratio is not in expected range
                           pairLevVec[pair.count]<-2;
                        }
   
                     } else {
                        # not find 1st isotope for C13
                        pairLevVec[pair.count]<-2;
                     }

                     if (isDebug==TRUE) 
                     { 
                        C13TagVec[pair.count]<-mzCCType;
                        intR1stIso_C12Vec[pair.count]<-mzACTypeList$intR; # intensity ratio of 1st isotope over C12
                     }
                     # indOri -> indexSample -> indexCurrentNo 
                     indLightVec[pair.count]<-indOri[indexSample[indexCurrentNo[m]]];               # mzA
                     indHeavyVec[pair.count]<-indOri[indexSample[indexCurrentNo[indexCReal]]];      # mzc
                     pairChargeVec[pair.count]<-mzACType;
                     pairTagVec[pair.count]<-mzATType;
                     
                     matched.peaks[indexCurrentNo[m]]<-TRUE;              # within current sample
                     matched.peaks[indexCurrentNo[indexCReal]]<-TRUE;     # within current sample
   
                     # remove all isotopes of C12, C13
                     if (intA>intSaturated)
                     {
                       indexIsoC12<-findIsotopes(mzVecCurrentNo[(m+1):nLenCurrentNo],intVecCurrentNo[(m+1):nLenCurrentNo],
                                                mzACType,mzATType,mzA,intA,mzTolIso,intRatioRangeChargeMatST[,mzACType],12); # after mzA
                     } else {
                       indexIsoC12<-findIsotopes(mzVecCurrentNo[(m+1):nLenCurrentNo],intVecCurrentNo[(m+1):nLenCurrentNo],
                                                mzACType,mzATType,mzA,intA,mzTolIso,intRatioRangeChargeMat[,mzACType],12); # after mzA
                     }
                     # label the isotopes matched[used]
                     if (length(indexIsoC12)==0)  { stop("Should not see this msg: C12 should have isotopes"); }
                     matched.peaks[indexCurrentNo[(indexIsoC12+m)]] <- TRUE;
                     if ( mzCCType>0)
                     {
                        if (intC>intSaturated)
                        {
                           indexIsoC13<-findIsotopes(mzVecCurrentNo[(indexCReal+1):nLenCurrentNo],intVecCurrentNo[(indexCReal+1):nLenCurrentNo],
                                                     mzCCType,mzATType,mzC,intC,mzTolIso,intRatioRangeChargeMatST[,mzCCType],13); #after mzC
                        }  else {
                           indexIsoC13<-findIsotopes(mzVecCurrentNo[(indexCReal+1):nLenCurrentNo],intVecCurrentNo[(indexCReal+1):nLenCurrentNo],
                                                     mzCCType,mzATType,mzC,intC,mzTolIso,intRatioRangeChargeMat[,mzCCType],13); #after mzC
                        }
                        # label the isotopes matched[used]
                        if (length(indexIsoC13)==0)  { stop("Should not see this msg: C13 should have isotopes"); }
                        #matched.peaks[indexSample[indexCurrentNo[(indexIsoC13+indexCReal)]]] <- TRUE;
                        matched.peaks[indexCurrentNo[(indexIsoC13+indexCReal)]] <- TRUE;
                     }
                       
                  } else {
                    next;
                  }
               } else {
                  # not in the four charge-types ==> lev 3 [so far], to find a pair?
                  # do nothing
                  next;   
               }
   
           }  # end of loop mzA
   
       }  # end of scan-number

       # fix lev3-peaks [set the chargeType,tagType]
       for (nL in 1:3)
       {
           indexIsNotMatch<-which(matched.peaks==FALSE);
           mzVecRef<-peaklist.sorted$mz[indLightVec[1:pair.count]]
           scanNumVecRef<-peaklist.sorted$No[indLightVec[1:pair.count]]
           oriLev3Peaks<-list(mz=mzVec[indexIsNotMatch],scanNum=scanNumVec[indexIsNotMatch])
           pkLev3<-fixLev3Peaks(oriLev3Peaks,mzTolLev3,mzVecRef,scanNumVecRef,pairChargeVec[1:pair.count],pairTagVec[1:pair.count],tagMat);
           numPKLev3<-pkLev3$numFix[1]
           if (numPKLev3>0)
           {
              print(paste("  findPeakPairsDupl.R:  lev3Pair found: ",numPKLev3," in Loop",nL,sep=""))
              tmpInd0 <- pair.count+1;
              tmpInd1 <- pair.count+numPKLev3;
              pairLevVec[tmpInd0:tmpInd1]<-3;
              indHeavyVec[tmpInd0:tmpInd1]<-indOri[indexSample[indexIsNotMatch[pkLev3$indFixH[1:numPKLev3]]]];
              indLightVec[tmpInd0:tmpInd1]<-indOri[indexSample[indexIsNotMatch[pkLev3$indFixL[1:numPKLev3]]]];
              pairChargeVec[tmpInd0:tmpInd1]<-pkLev3$chargeFix[1:numPKLev3]
              pairTagVec[tmpInd0:tmpInd1]<-pkLev3$tagFix[1:numPKLev3]
              matched.peaks[indexIsNotMatch[pkLev3$indexUsed[1:pkLev3$numUsed]]] <- TRUE;
              pair.count<-pair.count+numPKLev3;
           } else {
             break;
           }
       }
       # lev-4 pairs [distance based]
       indexIsNotMatch<-which(matched.peaks==FALSE);
       pkLev4<-list(scanNum=scanNumVec[indexIsNotMatch],mz=mzVec[indexIsNotMatch],int=intVec[indexIsNotMatch],oriIndex=indOri[indexSample[indexIsNotMatch]])
       pkLev4<-as.data.frame(pkLev4);
       pkLev4Sorted<-pkLev4[order(pkLev4$scanNum,pkLev4$mz), ]; rm(pkLev4)

       scanNumLev4<-pkLev4Sorted$scanNum;
       mzLev4 <-pkLev4Sorted$mz;
       intLev4<-pkLev4Sorted$int;
       scanNumLev4Uq<-unique(scanNumLev4);
       numScanNums0<-length(scanNumLev4Uq);
       for (nS0 in 1:numScanNums0)
       {
           scanNumC0<-scanNumLev4Uq[nS0];
           indexCurrentNoLev4<-which(scanNumLev4==scanNumC0);
           mz0List <- mzLev4[indexCurrentNoLev4];
           int0List<- intLev4[indexCurrentNoLev4];
           pkLev4New<-findPeakPairMZDist(mz0List,int0List);
           numPairLev4<-pkLev4New$numPair;
           if (numPairLev4>0)
           {
              tmpInd0 <- pair.count+1;
              tmpInd1 <- pair.count+numPairLev4;
              pairLevVec[tmpInd0:tmpInd1]<-4;
              indHeavyVec[tmpInd0:tmpInd1]<-pkLev4Sorted$oriIndex[indexCurrentNoLev4[pkLev4New$indexHeavy]];
              indLightVec[tmpInd0:tmpInd1]<-pkLev4Sorted$oriIndex[indexCurrentNoLev4[pkLev4New$indexLight]];
              pair.count<-pair.count+numPairLev4;
           } else {
             next;
           }
       }
 
   } # end of sample   
   
   # prepare the output
   print(paste("  findPeakPairsDupl.R:  Total pairs identified: ", pair.count, ".", sep=""));
   # create the return matrix
   # set up the pair matrix
   indLightVec<-indLightVec[1:pair.count]
   indHeavyVec<-indHeavyVec[1:pair.count]
   pairRatioVec<-pairRatioVec[1:pair.count]

   if (isDebug==TRUE)
   {
     if (isFT==TRUE)
     {
       pair.mat<-matrix(nrow=pair.count, ncol=18)
       colnames(pair.mat)<-c("sample","scanNum","rt","mz_light", "mz_heavy", "distance", "int_light", "int_heavy", "ratio_light_to_heavy","sn_light","sn_heavy","nCharge","nTag","pairLev","isSaturated","intRC13_C12_1stIso","nCharge_C13","intR_1stIso_C12");
     } else {
       pair.mat<-matrix(nrow=pair.count, ncol=16)
       colnames(pair.mat)<-c("sample","scanNum","rt","mz_light", "mz_heavy", "distance", "int_light", "int_heavy", "ratio_light_to_heavy","nCharge","nTag","pairLev","isSaturated","intRC13_C12_1stIso","nCharge_C13","intR_1stIso_C12");
     }
   } else {
     if (isFT==TRUE)
     {
        pair.mat<-matrix(nrow=pair.count, ncol=15)
        colnames(pair.mat)<-c("sample","scanNum","rt","mz_light", "mz_heavy", "distance", "int_light", "int_heavy", "ratio_light_to_heavy","sn_light","sn_heavy","nCharge","nTag","pairLev","isSaturated");
     } else {
       pair.mat<-matrix(nrow=pair.count, ncol=13)
       colnames(pair.mat)<-c("sample","scanNum","rt","mz_light", "mz_heavy", "distance", "int_light", "int_heavy", "ratio_light_to_heavy","nCharge","nTag","pairLev","isSaturated");
     }
   }
   pair.mat[ ,1]<-peaklist.sorted$sample[indLightVec]
   pair.mat[ ,2]<-peaklist.sorted$No[indLightVec]
   pair.mat[ ,3]<-peaklist.sorted$rt[indLightVec]
   pair.mat[ ,4]<-peaklist.sorted$mz[indLightVec]
   pair.mat[ ,5]<-peaklist.sorted$mz[indHeavyVec]
   pair.mat[ ,6]<-pair.mat[ ,5]-pair.mat[ ,4]
   pair.mat[ ,7]<-peaklist.sorted$maxo[indLightVec]
   pair.mat[ ,8]<-peaklist.sorted$maxo[indHeavyVec]
   pair.mat[ ,9]<-pair.mat[ ,7]/pair.mat[ ,8]
   # replace saturated peak ratio
   indSaturatedPeak<-which(pairRatioVec!=0)
   pair.mat[indSaturatedPeak,9]<-pairRatioVec[indSaturatedPeak]
   if (isFT==TRUE)
   {
      pair.mat[ ,10]<-peaklist.sorted$sn[indLightVec]
      pair.mat[ ,11]<-peaklist.sorted$sn[indHeavyVec]
      nL=2;
   } else {
      nL=0;
   }
   pair.mat[,(10+nL)]<-pairChargeVec[1:pair.count];
   pair.mat[,(11+nL)]<-pairTagVec[1:pair.count];
   pair.mat[, (12+nL)]<-pairLevVec[1:pair.count];

   # for isSaturated field
   isSaturated<-vector(mode="integer",length=pair.count)
   isSaturated[indSaturatedPeak]<-1
   pair.mat[, (13+nL)]<-isSaturated
   if (isDebug==TRUE)
   {
      pair.mat[ ,(14+nL)]<-intRC13_C12Vec[1:pair.count];
      pair.mat[ ,(15+nL)]<-C13TagVec[1:pair.count];
      pair.mat[ ,(16+nL)]<-intR1stIso_C12Vec[1:pair.count];
      write.table(pair.mat,file=debugfile,col.names=TRUE,sep=",",row.names=FALSE)
      pair.mat<-pair.mat[,1:(13+nL)]
   }
   # return the pair matrix
   return(pair.mat);
}

#TTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTT#
#######################################################################################
#function to find the charge-type
# Usage:
#      returnList<-findChargeType(mzlist,intlist,mzA,intA,mzDistVec,mzTolVec,intRatio,isUsed)
#                   mzlist: mz-list contains the mz to check the charge-type [within same scan-number]
#                  intlist: corresponding intensity list
#                      mzA: base mz
#                     intA: base intensity
#                mzDistVec: mz-dist vector for different charge types
#                           default: c(1.00336,0.50168,0.33445,0.25084);
#                       charge type:      1        2      3       4
#                 mzTolVec: mz-tolerance vector different charge types
#                           default: c(2.5,2.5,2.5,2.5)*1e-6 ==> 2.5ppm
#         intRatioRangeMat: intensity ratio for a valid charge type
#                           default: matrix(c(0.15,0.75,          chargetype-1
#                                             0.15,0.75,          chargetype-2
#                                             0.15,0.75,          chargetype-3
#                                             0.15,0.75),2,4);    chargetype-4
#                            different columns ==> different chargeType
#            return a list:
#                           chargeType=chargeTypeFound,ind=nCurrent,mz=mzB,int=intB
# Update:
#     Mar 6, 2012, recoded as a function
#     Mar24, 2012, re-checked, keep the charge-type consistent
########################################################################################
findChargeType<-function(mzlist,intlist,mzA,intA,mzDistVec, mzTolVec,intRatioRangeMat,isUsed)
{
    isInFourType<-FALSE;
    chargeTypeFound<-0;
    nChargeType<-length(mzDistVec);
    numRow<-length(mzlist);
    maxMZCharge<-max(mzDistVec)
    nCurrent<-0; # maybe useful later

    # search the more charge first
    for (nC in nChargeType:1)
    {
       for (n in 1:numRow)
       {
           if (isUsed[n]) {next;};
           #  check charge-type
           mzB <- mzlist[n];
           dist <- mzB-mzA;
           if ( dist < 0 ) 
           { 
              print("  Negative mass difference in function: findChargeType [findPeakPairsDupl.R]");
              stop(paste("  should not see the above msg: mzA=",mzA," mzB=",mzB,sep=""));
           }
           # *********************    Double Check Here   *********************************************
           if (abs(dist-mzDistVec[nC]) < (mzTolVec[nC]*mzA) )
           {
              intB<-intlist[n];
              intRatio<- intB/intA;
              if ( intRatio > intRatioRangeMat[1,nC] && intRatio<intRatioRangeMat[2,nC] )
              {
                 isInFourType<- TRUE;
                 chargeTypeFound<- nC;
                 nCurrent<-n;
                 break;
              }
           }
           if (dist>2*maxMZCharge) { break; }  ### NOTE HERE
       }
       if (isInFourType==TRUE) { break; }  ### only detect the first possible charge-type
   }
   # return the type of Charge
   
   if (isInFourType==TRUE)
   {
      return(list(chargeType=chargeTypeFound,ind=nCurrent,mz=mzB,int=intB,intR=intRatio))
   } else {
      return(list(chargeType=chargeTypeFound))
   }
}

#TTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTT#
#######################################################################################
# function to find the tag-type
# Usage:
#       returnList<-findTagType(mzlist,intlist,mzA,intA,mzDistVec,mzTolVec,mzB,intB,intRatioC12_1stTag,intRatioC13_to_1stISO)
#                   mzlist: mz-list contains the mz to check the charge-type [within same scan-number]
#                  intlist: corresponding intensity list
#                      mzA: base mz
#                     intA: base intensity
#                mzDistVec: mz-dist vector for different charge types
#                           default: tagMat[,nChargeType]
#                           tagMat<-matrix(c(2.0067,4.0133,6.02001,8.02668,        # charge-type 1
#                                            1.0034,2.0067,3.01008,4.01344,        # charge-type 2
#                                            0.6689,1.3378,2.00670,2.67560,        # charge-type 3
#                                            0.5017,1.0034,1.50504,2.00672),4,4);  # charge-type 4
#
#                 mzTolVec: mz-tolerance vector different charge types
#                           default: c(2.5,2.5,2.5,2.5)*1e-6  ==> [ppm]
#                      mzB: mz of 1st isotope
#                     intB: intensity of 1st isotope
#       intRatioC12_1stTag: the "max" intensity ratio of C12 over possible tag 1 [for test, 3 is used]
#    intRatioC13_to_1stISO: intensity ratio between C13 to 1st isotope
#                           default: c(0.5, 0.5,0.5,0.5) [max ratio otherwise ==> check next tag, and possible a pair[those C12,C13]]
#            return a list:
#                           tagType, ind, mz, int
#######################################################################################
findTagType<-function(mzlist,intlist,mzA,intA,mzDistVec,mzTolVec,mzB,intB,intRatioC12_1stTag,intRatioC13_to_1stISO,isUsed)
{
   nSearchStart<-1;
   numRow<-length(mzlist)
   nTagType<-length(mzDistVec);
   indTag<-vector(mode='integer',length=nTagType);
   mzCVec<-vector(mode='numeric',length=nTagType);
   intCVec<-vector(mode='numeric',length=nTagType);
   tagTypeFound<-0;

   isTagTypeFound<-FALSE;
   isPreTag1Found<-FALSE;
   for (nTag in 1:nTagType)
   {
      for (n in nSearchStart:numRow)
      {
          if (isUsed[n]) {next;};
          mzC <- mzlist[n];
          dist <- mzC-mzA;
          if (dist > (mzDistVec[nTagType]*1.1)) { break; }
          if ( dist < 0 ) { stop("should not see this msg2: negative mass difference in findTagType [findPeakPairsDupl.R]!!!"); }
          # ***********************     DOUBLE CHECK  HERE     ********************************************* 
          if ( abs(dist-mzDistVec[nTag]) < (mzTolVec[nTag]*mzA) )
          {
             intC<-intlist[n];
             if ((intC/intB) > intRatioC13_to_1stISO[nTag])
             {
                if (nTag==1)
                {
                   # check possible next tag
                   if (intA/intC>intRatioC12_1stTag)
                   {
                      isPreTag1Found<-TRUE
                      intTag1<-intC
                      mzTag1<-mzC
                      indTag1<-n
                      break;
                   } else {
                     isTagTypeFound<-TRUE
                     tagTypeFound<-nTag;
                     indTag[nTag]<-n;
                     mzCVec[nTag]<-mzC;
                     intCVec[nTag]<-intC;
                     break;
                   }
                } else if (nTag==2) {
                  if (isPreTag1Found==TRUE)
                  {
                     # tag-2 found
                    if (intA/intC>intRatioC12_1stTag)
                    {
                      # keep tag1
                      isTagTypeFound<-TRUE
                      tagTypeFound<-1;
                      indTag[1]<-indTag1;
                      mzCVec[1]<-mzTag1;
                      intCVec[1]<-intTag1;
                      break;
                    } else {
                      # use tag2
                      isTagTypeFound<-TRUE
                      tagTypeFound<-2;
                      indTag[nTag]<-n;
                      mzCVec[nTag]<-mzC;
                      intCVec[nTag]<-intC;
                      break;
                    }
                  } else {
                    # tag2, no tag1
                    isTagTypeFound<-TRUE
                    tagTypeFound<-2;
                    indTag[nTag]<-n;
                    mzCVec[nTag]<-mzC;
                    intCVec[nTag]<-intC;
                    break;
                  } 
                } else {
                 if (isPreTag1Found==TRUE) {
                    # tag-2 is not found, keep tag 1
                    isTagTypeFound<-TRUE
                    tagTypeFound<-1;
                    indTag[1]<-indTag1;
                    mzCVec[1]<-mzTag1;
                    intCVec[1]<-intTag1;
                    break;
                 } else {
                    isTagTypeFound<-TRUE
                    tagTypeFound<-nTag;
                    indTag[nTag]<-n;
                    mzCVec[nTag]<-mzC;
                    intCVec[nTag]<-intC;
                    break;
                 }
               } # if condition for different tag types
             } else {
               # if lower-intensity-kind-isotopic is found, discard it, go for next possible tag-type 
               nSearchStart<-n+1;
               break; # could enable this line
             } # match intensity
          } # match mz
      } # loop for all mz
      if (isTagTypeFound) { break; }  # only previous tag is found, will check further
      if (nSearchStart>numRow) { break;}
   } # loop for all tag-types
   if (isPreTag1Found==TRUE && isTagTypeFound==FALSE)
   {
      tagTypeFound<-1;
      nTag<-1
      indTag[nTag]<-indTag1;
      mzCVec[nTag]<-mzTag1;
      intCVec[nTag]<-intTag1;
   }
   # 
   # return
   return(list(tagType=tagTypeFound,ind=indTag,mz=mzCVec,int=intCVec))
}
#TTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTT#
#######################################################################################
# function to find the index of isotopes of given mzA
# Usage:
#      indexdOfIsotope <- findIsotopes(mzlist,intlist,nCharge,nTag,mzA,intA,mz_tol,nOpt)
#                      mzlist: the vector contains all the mz used for isotope search,
#                              assuming that input mzlist is sorted ascending by mz already
#                     intlist: intensity list
#                     nCharge: number of Charge, related to mz-step for isotopes
#                        nTag: number of tag,  related to number of possible isotopes  
#                         mzA: base mz
#                      mz_tol: mz tolerance used to detect isotopes
#                        nOpt: 12 for C12, and 13 for C13, related to the number of isotopes
#                        NOTE: intRatio [0.15 0.75], <0.5, <0.25, <0.125
# Update:
#           Mar 6, 2012: first coded
#           Mar 7, 2012: add the intensity ratio control, assuming always decrease for isotopes
#                        if a "larger" intensity is found, will stop searching further isotopes
#           Mar31, 2012: use different mzTol for different charget type
#                        disabled the break [when distance matches but ratio does not]
#           Apr05, 2012: consider multiple isotopes for each nLev
#                        increase search range from 1.1 to 1.2

findIsotopes<-function(mzlist, intlist, nCharge, nTag, mzA, intA, mz_tol, intR1st, nOpt)
{
    intRatio <- c(NA,0.5,0.25,0.125);   # intensity ratio for rest isotopes
    isDebug<-FALSE;
    if (nOpt!=12 && nOpt!=13)
    {
       stop(paste("Not defined nOpt in findIsotopes: nOpt="+nOpt))
    }

    # distance setting for isotopes
    # create the mz-dist-array for isotopes 
    isoMZBase<-matrix(rep(c(1.00336,0.50168,0.33445,0.25084),4),4,4); # row: ChargeType, col: [nLev of isotope] tagType
    for (nc in 2:ncol(isoMZBase))
    {
        isoMZBase[,nc]<-isoMZBase[,nc]*nc;
    }

    if (nOpt==12)
    {
       isoLevForTag<-t(matrix(c(1,3,4,4,1,3,4,4,1,3,4,4,1,3,4,4),4,4));
    } else {
       isoLevForTag<-t(matrix(c(2,4,4,4,2,4,4,4,2,4,4,4,2,4,4,4),4,4));
    }
    
    indISO <- vector(mode='integer',length=8); 
     nRows <- length(mzlist);
    nStart <- 1;

    # compute special mz-window for dist @ 2nd isotope
    numISO<-0;
    for (nLev in 1:isoLevForTag[nCharge,nTag])
    {
        if (isDebug==TRUE) {numLocal<-0;}
        isFoundISO<-FALSE;
        for (nr in 1:nRows)
        {
            mzB <- mzlist[nr];
            dist <- mzB-mzA;
            if ( dist < 0 ) { stop("Should not see this msg in findIsotopes [findPeakPairsDupl.R]: negative distance is found!"); }
            if ( dist > (2*isoMZBase[nCharge,nLev])) { break; }

            if ( abs(dist-isoMZBase[nCharge,nLev]) < (mz_tol[nCharge]*mzA) )
            {
               if (nLev==1)
               {
                  rr<-intlist[nr]/intA;
                  if (rr>intR1st[1] && rr<intR1st[2])
                  {
                     # valid isotopes
                     isFoundISO<-TRUE;
                     numISO<-numISO+1;
                     indISO[numISO]<-nr;
                     intISO0<-intlist[nr];
                     if (isDebug==TRUE) {numLocal<-numLocal+1;}
                  }
               } else {
                  rr<-intlist[nr]/intISO0;
                  if (rr < intRatio[nLev])
                  {
                     # valid isotopes
                     isFoundISO<-TRUE;
                     numISO<-numISO+1;
                     indISO[numISO]<-nr;
                     intISO0<-intlist[nr];
                     if (isDebug==TRUE) {numLocal<-numLocal+1;}
                  }
               }
               if (isDebug==TRUE && numLocal>1) 
               {
                  print(paste("findIsotopes: Found two or more isotopes: mzA=",mzA," numLocal=",numLocal,sep=""));
                  print(paste("              if possible, please reduce the mzTolIso when nCharge=",nCharge));
               }
            }
        } # loop for mzB
        if (!isFoundISO) { break; }
        if (nStart>nRows){ break; }
    } # loop for different isotope-types

    # return the index of isotopes
    return(indISO[indISO>0])
}
#TTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTT#
#######################################################################################
# The function finds all pairs of peaks from a peak table based on the given m/z distance
# Usage:
#        returnList<-findPeakPairMZDist(mzlist,intlist)
#              mzlist: mz-list contains the mz to find pairs [within same scan-number,sorted ascending assumed]
#             intlist: corresponding intensity list
#        return a list:
#              numPair,indexLight,indexHeavy, pairDist, pairRatio
#
# Origin:         Jeff Xia, June, 2009, University of Alberta
# Update:
#    Sep  , 2011  *recode the distance conditions
#    Jan29, 2012  *optimized and scan number based.
#    Jan31, 2012  *optimized again with vectorized index [speed up ~10%]
#    Mar07, 2012  *simplified for a single scan-number
#    Mar31, 2012  *only dist-2 type is considered
####################################################

findPeakPairMZDist<-function(mzlist,intlist)
{
  numRowC<-length(mzlist);
  maxLen<-floor(numRowC/2);
  indLightVec<-vector(mode="integer",length=maxLen);
  indHeavyVec<-vector(mode="integer",length=maxLen);
  pairDistVec<-vector(mode="numeric",length=maxLen);
  pairRatioVec<-vector(mode="numeric",length=maxLen);
  
  # set up a vector to remember peaks already matched
  matched.peaks<-vector(mode='logical', length=numRowC);
  m <- 0;
  pair.count = 0;

  # iterate through all peaks, the lighter one should not be the last one if sorted, thus m<numRowC-1
  while ( m < (numRowC-1) ) 
  {
        m <- m+1;
        mzA<-mzlist[m];
        # if matched with some peak already, check next one
        if (matched.peaks[m]) { next; }

        for (n in (m+1):numRowC)
        {
             # should not matched with some others even with same scan number
             if (matched.peaks[n]) { next; }

             mzB<-mzlist[n];
             if (mzB<mzA)
             {
                stop("should not see this msg: negative mass difference in findPeakPairMZDist!!!")
             }
             dist <-mzB-mzA;
             if ( dist >= 2.0047 & dist <= 2.0087)
             {
                intB<-intlist[n]
                pair.count<-pair.count+1;
                indLightVec[pair.count]<-m
                indHeavyVec[pair.count]<-n;
                matched.peaks[n]<-TRUE;
                pairDistVec[pair.count]<-dist;   #fixed Jan8, 2013 by Li Ronghong
                pairRatioVec[pair.count]<-intB/intlist[m];
                break;
             }
             if (dist>3) {break;} # no further searching

        } # loop of n
  } # loop of m
  
  if (pair.count>0)
  {
     indexHeavy<-indHeavyVec[1:pair.count]
     indexLight<-indLightVec[1:pair.count]
     pairDist<-pairDistVec[1:pair.count]
     pairRatio<-pairRatioVec[1:pair.count]
  } else {
     indexHeavy<-0;
     indexLight<-0;
     pairDist<-0;
     pairRatio<-0;
  }
  # return the list
  return(list(numPair=pair.count,indexLight=indexLight, indexHeavy=indexHeavy,pairDist=pairDist,pairRatio=pairRatio))
}

#TTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTT#
#######################################################################################
# to find chargeType and tagType for Lev3-peaks
# Usage:
#     newPK <-fixLev3Peaks(oldPKList,mz_tol,mzVecRef,scanNumVecRef,chargeTypeRef,tagTypeRef,tagMat)
#######################################################################################

fixLev3Peaks<-function(pkLev3,mz_tol,mzVecRef,scanNumVecRef,chargeTypeRef,tagTypeRef,tagMat)
{
     #order scan-number, mz
     pkLev3<-as.data.frame(pkLev3);
     numRow<-nrow(pkLev3);
     pkLev3$oriIndex<-c(1:numRow);

     pkLev3Sorted<-pkLev3[order(pkLev3$scanNum,pkLev3$mz), ]; rm(pkLev3)
     isUsed<-vector(mode='logical',length=numRow);

     indFixedLight<-vector(mode='integer',length=numRow);
     indFixedHeavy<-vector(mode='integer',length=numRow);
     indUsed<-vector(mode='integer',length=numRow);
     chargeFixded<-vector(mode='integer',length=numRow);
     tagFixed<-vector(mode='integer',length=numRow);

     scanNumLev3<-pkLev3Sorted$scanNum;
     mzLev3<-pkLev3Sorted$mz;
     oriIndLev3<-pkLev3Sorted$oriIndex;

     scanNumLev3Uq<-unique(scanNumLev3);
     numScanNums<-length(scanNumLev3Uq);
     numFix<-0;
     numUsed<-0;
     maxSearch<-3; # maximum scan number search

     for (nS in 1:numScanNums)
     {
         scanNumC<-scanNumLev3Uq[nS];
         #print(paste("Checking scanNum:",scanNumC,sep=""))
         indexCurrentNo<-which(scanNumLev3==scanNumC);
         mzLev3CurrentNo<-mzLev3[indexCurrentNo]
         nLenCurrentNo<-length(indexCurrentNo)
         if (nLenCurrentNo<=1) { next; }
         for (m in 1:(nLenCurrentNo-1))
         {
             if (isUsed[indexCurrentNo[m]]==TRUE) 
             {
                next;
             }
             mzCurrent<-mzLev3CurrentNo[m]
             indtmp<-which(abs(mzVecRef-mzCurrent)<mzCurrent*mz_tol & abs(scanNumVecRef-scanNumC)<=maxSearch)
             if (length(indtmp)>0)
             {
                # choose the closest one [scanNum] 
                if (length(indtmp)>1)
                {
                   scanNumTmpDiff<-abs(scanNumVecRef[indtmp]-scanNumC)
                   minNumDiff<-min(scanNumTmpDiff)
                   indMin<-which(scanNumTmpDiff==minNumDiff)[1]
                   indtmp<-indtmp[indMin]
                }
                # get chargeType, tagType
                chargeTypetmp<-chargeTypeRef[indtmp];
                tagTypetmp<-tagTypeRef[indtmp]
                # search C13 with given chargeType and tagType in mzLev3CurrentNo [conside intRatio or not]
                C13MZDist<-tagMat[tagTypetmp,chargeTypetmp];
                nCurrent<-m+1
                for (m2 in nCurrent:nLenCurrentNo)
                {
                    mzC<- mzLev3CurrentNo[m2]
                    dist<- mzC-mzCurrent
                    if (dist<0) 
                    {
                       print("Negative distance in function: fixLev3Peaks [findPeakPairsDupl.R]!")
                       stop("Should not see the above msg!")
                    }
                    if (abs(dist-C13MZDist) < mz_tol*mzCurrent)
                    {
                       # if found ==> numFix<-numFix+1
                       isUsed[indexCurrentNo[m2]]<-TRUE; # C13
                       numFix<-numFix+1;
                       indFixedLight[numFix]<-oriIndLev3[indexCurrentNo[m]];
                       indFixedHeavy[numFix]<-oriIndLev3[indexCurrentNo[m2]];
                       chargeFixded[numFix]<-chargeTypetmp;
                       tagFixed[numFix]<-tagTypetmp;
                       numUsed<-numUsed+1;
                       indUsed[numUsed]<-oriIndLev3[indexCurrentNo[m2]];
                       break;
                    }
                }
                isUsed[indexCurrentNo[m]]<-TRUE;
                numUsed<-numUsed+1;
                indUsed[numUsed]<-oriIndLev3[indexCurrentNo[m]];
               # no touch for later distance-based pairing 
             } else {
             } # end check-close
         } # end loop of each lev3-peak
     } # end loop of each scan-number
     
     # return
     if (numFix>0)
     {
        return(list(numFix=numFix,numUsed=numUsed,indFixH=indFixedHeavy[1:numFix],indFixL=indFixedLight[1:numFix],chargeFix=chargeFixded[1:numFix],tagFix=tagFixed[1:numFix],indexUsed=indUsed[1:numUsed]))
     } else {
        return(list(numFix=numFix,numUsed=numUsed))
     }
}
