# group pairs with same charge number and tag number
#
# Usage:
#     newPK <-groupPairs(oldPairList,mzTol, intRatioPeakOverValley)
#                   mzTol                   : mz tolerance used to define a mz-group
#                   intRatioPeakOverValley  : the ratio of a good peak intensity over that of the valley
# Update:
#     Apr01, 2012:  done first version
#     Apr04, 2012:  add debug file
#     Apr05, 2012:  add swith for findGoodPeaks[ simple way or not]

groupPairs<-function(pkList,mzTol=5e-6,intRatioPeakOverValley=2,badSpectraFile=NA,debugFile=NA)
{

     isDebug<-FALSE;
     if (length(debugFile)!=0 && !is.na(debugFile))
     {
        isDebug<-TRUE;
     }

     #order charge number, then tag number
     pkList<-as.data.frame(pkList);
     numRow<-nrow(pkList);

     pkListSorted<-pkList[order(pkList$nCharge,pkList$nTag), ]; rm(pkList)
     isGoodPairs<-vector(mode='logical',length=numRow);
     if (isDebug==TRUE)
     {
        mzIDVec<-vector(mode='integer',length=numRow);
        noIDVec<-vector(mode='integer',length=numRow);
     }
     
     chargeVec<-pkListSorted$nCharge;
     tagVec<-pkListSorted$nTag;
     intVec<-pkListSorted$int_light;
     mzVec<-pkListSorted$mz_light;

     scanNumVec<-pkListSorted$scanNum;
     oriIndex<-c(1:numRow);
     chargeUq<-unique(chargeVec);
     tagUq<-unique(tagVec);
     numTolChargeType<-length(chargeUq);
     numTolTagType<-length(tagUq);
     maxScanMissing<-3
     isSimpleFindPeaks<-TRUE;

     # read bad spectra if need
     if (is.na(badSpectraFile))
     {
        badSpectra<-NA;
     } else {
       badSpectra<-scan(badSpectraFile,quiet=TRUE);
     }
     #
     for (nC in 1:numTolChargeType)
     {
         for (nT in 1:numTolTagType)
         {
             if (chargeUq[nC]!=0 && tagUq[nT]==0) { next; }
             if (chargeUq[nC]==0 && tagUq[nT]!=0) { break; }
             # find pairs with same chargeType and tagType
             indexSCT<-which(chargeVec==chargeUq[nC] & tagVec==tagUq[nT]);
             if (length(indexSCT)<1) { next;}
             mzCurrent<-mzVec[indexSCT];
             # label pairs group with close mzs
             mzID<-labelMZGroup(mzCurrent,mzTol);
             numGroup<-length(unique(mzID)); 
             for (nG in 1:numGroup)
             {
                 indexSameMZ<-which(mzID==nG)
                 intCgroup <- intVec[indexSCT[indexSameMZ]]
                 intCMax <- max(intCgroup)
                 numMZ <- length(indexSameMZ)

                 if (isDebug) { mzIDVec[indexSCT[indexSameMZ]]<-nG; }
                 if (numMZ==1) 
                 { 
                    isGoodPairs[indexSCT[indexSameMZ]]<-FALSE; 
                    if (isDebug) { noIDVec[indexSCT[indexSameMZ]]<--1; }
                    next;
                  }

                 #sort by scan number
                 scanNumCgroup<-scanNumVec[indexSCT[indexSameMZ]]
                 ordIndex<-order(scanNumCgroup)
                 scanNumCgroupSorted<-scanNumCgroup[ordIndex]
                 intCgroupSorted <- intCgroup[ordIndex]

                 scanNumID<-labelScanNumID(scanNumCgroupSorted,maxScanMissing,badSpectra);
                 numNoGroup<-length(unique(scanNumID)); 
                 numNoNum<-length(scanNumID); 
                 if (numNoNum>1)
                 {
                    for (nScanNum in 1:numNoGroup)
                    {
                        indNoID<-which(scanNumID==nScanNum)
                        if (isSimpleFindPeaks==TRUE)
                        {
                           tmpIntVect<-intCgroupSorted[indNoID];
                           indGoodLocal<-which(tmpIntVect==max(tmpIntVect))[1]
                        } else {
                           indGoodLocal<-findGoodPeaks(intCgroupSorted[indNoID],intRatioPeakOverValley);
                        }
                        isGoodPairs[indexSCT[indexSameMZ[ordIndex[indNoID[indGoodLocal]]]]]<-TRUE;
                        if (isDebug) { noIDVec[indexSCT[indexSameMZ[ordIndex[indNoID]]]]<-nScanNum; }
                    } # scan-number group
                 }
             } # mz-group
         } # tag-type
     } #  charge-type


     if (isDebug==TRUE)
     {
        pkListSorted$mzID<-mzIDVec;
        pkListSorted$noID<-noIDVec;
        pkListSorted$isKeep<-isGoodPairs;
        write.table(pkListSorted,file=debugFile,col.name=TRUE,sep=",",row.name=FALSE);
        pkListSorted$mzID<-NULL;
        pkListSorted$noID<-NULL;
        pkListSorted$isKeep<-NULL;
     }
     # return
     numPairRemove<-length(isGoodPairs==FALSE)
     print(paste("groupPairs.R: a total of  ",numPairRemove," pairs removed!"),sep="")
     return(pkListSorted[which(isGoodPairs==TRUE),]);
}

labelMZGroup<-function(mzList,mzTol)
{
  # find the unique MZs with a given mzTol
  ordIndex<-order(mzList); # sort mz ascending
  mzListSorted<-mzList[ordIndex]; rm(mzList)
  numMZ<-length(mzListSorted);
  mzGroupID<-vector(mode='integer',length=numMZ); 
  isUsed<-vector(mode='logical',length=numMZ); 

  numMZGroup<-0;
  for (n in c(1:numMZ))
  {
     if (isUsed[n]==FALSE)
     {
         mzC<-mzListSorted[n]
         numMZGroup<-numMZGroup+1;
         indCloser<-which(mzListSorted>=mzC & (mzListSorted-mzC)<(mzC*mzTol));
         mzGroupID[ordIndex[indCloser]]<-numMZGroup;
         isUsed[indCloser]<-TRUE;
     }
  }
  return(mzGroupID);
}

labelScanNumID<-function(scanNumList,maxScanMissing,badSpectra=NA)
{
   noID<-1;
   numNo<-length(scanNumList);
   noGroupID<-vector(mode='integer',length=numNo); noGroupID[]<-noID;
   if (is.na(badSpectra[1]))
   {
      #no bad spectra
      m<-1;
      while (m < numNo)
      {
          if ((scanNumList[m+1]-scanNumList[m])<=maxScanMissing)
          {
             noGroupID[m+1]<-noID;
          } else {
             noID<-noID+1;
             noGroupID[m+1]<-noID;
          }
          m<-m+1;
      }
   } else {
     badSpectra<-sort(badSpectra);
     m<-1;
     while (m<numNo)
     {
        indBad<-which(badSpectra<scanNumList[m+1] & badSpectra>scanNumList[m])
        if (length(indBad)>0)
        {
           noDiff<-max(scanNumList[m+1]-max(badSpectra[indBad]),min(badSpectra[indBad])-scanNumList[m]);
        } else {
           noDiff<-scanNumList[m+1]-scanNumList[m];
        }
        if (noDiff<=maxScanMissing)
        {
           noGroupID[m+1]<-noID;
        } else {
           noID<-noID+1;
           noGroupID[m+1]<-noID;
        }
        m<-m+1;
     }
   } 
   return(noGroupID);
}

findGoodPeaks<-function(intList,minIntRatio=2)
{
  # need return the local index of good peaks [the highest one must be labeled as good peak]
  # assume input are sorted by scan number ascending already, and should be continuous

  numInt<-length(intList);
  indexPeak<-vector(mode='integer',length=numInt);
  numPeak<-0;
  if ( intList[1] > intList[2] ) 
  { 
     numPeak<-numPeak+1;
     indexPeak[numPeak]<-1;
  }

  if ( intList[numInt] > intList[(numInt-1)] ) 
  { 
     numPeak<-numPeak+1;
     indexPeak[numPeak]<-numInt;
  }
  maxInt<-max(intList);
  indMax<-which(intList==maxInt)[1];
  if (indMax!=1 && indMax!=numInt)
  {
     numPeak<-numPeak+1;
     indexPeak[numPeak]<-indMax;
  }

  indPreV<-1;
  isPeakFound<-FALSE;
  # 1st half
  indP<-1;
  if (indMax > 2)
  {
     for (n in 2:(indMax-1) )
     {
         if ( (intList[n] > intList[(n-1)]) && (intList[(n+1)]<intList[n]))
         {
            indP<-n;
         } else if ( (intList[n]<intList[(n-1)]) && (intList[(n+1)]>intList[n]) ) {
            indV<-n;
            if ( ((indV-indP)>1) && ((intList[indP]/intList[indV])>minIntRatio) )
            {
               numPeak<-numPeak+1;
               indexPeak[numPeak]<-indP;
            }
         }
     }
  }
  # 2nd half
  if (indMax< (numInt-1))
  {
     for ( n in (indMax+1):(numInt-1) )
     {
         if ( (intList[n]< intList[(n-1)]) && (intList[(n+1)] > intList[n]) )
         {
            indV<-n;
         } else if ( (intList[n]>intList[(n-1)]) && (intList[(n+1)]<intList[n])) {
            indP<-n;
            if ( ((indP-indV)>1) && ((intList[indP]/intList[indV]) > minIntRatio ) )
            {
               numPeak<-numPeak+1;
               indexPeak[numPeak]<-n;
            }
         }
     }
  }
  # return
  return(unique(indexPeak[1:numPeak]))
}
